// Super Cell Alarm
// Copyright (c) 2016, 2017, 2018, 2019, 2020 by Haute Solutions / Ed Van Every (www.haute-solutions.com)
//
// FUNCTION: This sketch will implement the SuperCell Alarm functionality on the associated hardware.
//           This is a full featured Vehicle Alarm System providing tradional Alarm Alerts (Lights, Siren) as well
//           as SMS notification with GPS tracking.  
//           (Please refer to the Sleeper Cell Alarm Documentation for further information)
//
// HARDWARE: This sketch requires the following Hardware:
//
//           Haute Solutions MeGGaBoard (2G/3G) with Haute Solutions SuperCell Alarm Shield (V1.0)
//           -or-
//           Haute Solutions ESP324G with Haute Solutions SuperCell Alarm Shield (V2.0)
//
// SOFTWARE: This sketch requires the following Libraries and Software:
//
//           Haute Solutions GSMSMS Library
//
// LICENSE:  This program is free software: you can redistribute it and/or modify
//           it under the terms of the GNU General Public License as published by
//           the Free Software Foundation, either version 3 of the License, or
//           (at your option) any later version.
//
//           This program is distributed in the hope that it will be useful,
//           but WITHOUT ANY WARRANTY; without even the implied warranty of
//           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//           GNU General Public License for more details.
//           <http://www.gnu.org/licenses/>.
//
// ----------------------------- Sleeper Cell Alarm -----------------------------------------------------------------------------------------------

#define VERSION "V2.00"                     // Firmware Version Info

#define GOOGLE_MAPS_LOCATOR "https://www.google.com/maps/search/?api=1&query="

// Persistant Storage Info (EEPROM Info): Character Data, NULL Terminated
// All Fields are 20 Chars max (Inc NULL Terminator)
// The values below are the Field Offsets into EEPROM memory for start of each string
// Currently (V2 EEPROM Layout) Uses 580 Bytes Total

#define EESIG           0                   // EEPROM Data Valid Signature ("SUPERCELL EEPROM V1")
#define EEALARMID       1                   // Alarm ID ("Trailer") - For Multiple Alarms
#define EEPWD           2                   // Alarm Password
#define EEUSEGPS        3                   // GPS Function: {"0"=OFF, "1"=ON}
#define EEARMED         4                   // Persistant Arming State (For Power Up): {"0"=OFF, "1"=ON}
#define EEARMDELAY      5                   // Seconds Delay Before Arming After Power On (Allow Sensors to Stabilize)
#define EESMSINTERVAL   6                   // SMS Send Interval {"#" of Minutes}
#define EEVMIN          7                   // Voltage Alert Threshold (approx)
#define EET1NAME        8                   // Trigger 1 Name {"Door", "Motion", etc)
#define EET1INFO        9                   // Trigger 1 Config Info (See Below)
#define EET2NAME       10                   // Trigger 2 Name {"Door", "Motion", etc)
#define EET2INFO       11                   // Trigger 2 Config Info (See Below)
#define EET3NAME       12                   // Trigger 3 Name {"Door", "Motion", etc)
#define EET3INFO       13                   // Trigger 3 Config Info (See Below)
#define EER1ACTION     14                   // Output Relay 1 Action {"0"=OFF, "1"=ON, "2"=FLASH}
#define EER1NAME       15                   // Output Relay 1 Name {"Siren", "Lights", "Brakes", etc)
#define EER2ACTION     16                   // Output Relay 2 Action {"0"=OFF, "1"=ON, "2"=FLASH}
#define EER2NAME       17                   // Output Relay 2 Name {"Siren", "Lights", "Brakes", etc)
#define EER3ACTION     18                   // Output Relay 3 Action {"0"=OFF, "1"=ON, "2"=FLASH}
#define EER3NAME       19                   // Output Relay 3 Name {"Siren", "Lights", "Brakes", etc)
#define EER4ACTION     20                   // Output Relay 4 Action {"0"=OFF, "1"=ON, "2"=FLASH}
#define EER4NAME       21                   // Output Relay 4 Name {"Siren", "Lights", "Brakes", etc)
#define EECHIRP        22                   // Outputs to "Chirp" on Arm/Disarm
#define EENOTIFY       23                   // Start of SMS Notification List

#define EEPROM_BYTES  4096                  // Mega2560 Has 4096 Bytes of EEPROM

#define EEPROMSig "SUPERCELL V2.1"          // EEPROM Valid Signature (Version is EEPROM specific - NOT Firmware Specific)
#define EEFIELDSIZE    20                   // Size of each field/string (inc Terminator)
#define EEMAXSMS        5                   // Max Number of SMS Notification Numbers

#define SMSDELAY        2                   // Number of seconds to wait between SMS Messages (to different numbers)
#define CHIRPTIME       100                 // Number of milliseconds for Chirp Pulse

// EEPROM Trigger Configuration Fields (20 Bytes)
//
// Example: "1,1234,15"
// Field 1 = Priority (0-4)
// Field 2 = Outputs to Trigger (1234, 134, 23, 0, etc)
// Field 3 = Reset Time {minutes}

// Trigger Jumpers on SuperCell Alarm Shield
// NEG Trigger = Jump 1-2, 3-4
// POS Trigger = Jump 2-3, 4-5

#define NUMCMDS 28
const char *CmdArray[NUMCMDS] = { 
  "ALARMID:",
  "USEGPS:",
  "SMSINTERVAL:",
  "ARMDELAY",
  "TRIGGER:",
  "SHOWTRIGGERS",
  "OUTPUT:",
  "SHOWOUTPUTS",
  "CHIRP:",
  "NOTIFY:",
  "SHOWNOTIFY",
  "CLEARNOTIFY",
  "ARM",
  "DISARM",
  "RESET",
  "GPSLOC",
  "STATUS",
  "NETINFO",
  "CONFIG",
  "HELP",
  "PWD:",
  "VMIN:",
  "CLEAREEPROM",     //Undocumented
  "SIMINFO",         //Undocumented
  "MEINFO",          //Undocumented
  "CELLINFO",        //Undocumented
  "OPERATORS",       //Undocumented
  "ATCMD:"           //Undocumented
};

const char *BadCmd="Bad Command! (Try \"HELP\"!)";
const char MsgEnd[2]={0x1a,'\0'};                     // Chars to Signal End of Message

typedef struct {
  int led;
  unsigned long msdelay;
  unsigned long timeref;
  int blinks;
  int state;
} BLINKSTRUCT;

int RTBlinker(BLINKSTRUCT *blinkstruct, int blinkcount = -3, unsigned long msdelay = 0);

#define BUFFSIZE 300
#define NUMSIZE 20

#define TMULTIPLIER 60000   // Time Multiplier (Normally 60000 to convert millis to minutes)

// --------------------------------------------------------------------------------------------------------------------------------------------------

#include <EEPROM.h>
#include <GSMSMS.h>                      // GSM SMS Library

// Prep for ESP324G board
#if defined(ESP32)
#define T1 36          // Trigger 1 (White Wire)
#define T2 39          // Trigger 2 (Grey Wire)
#define T3 34          // Trigger 3 (Purple Wire)
#define AD 35          // Arm/Disarm (Blue Wire)
#define R1 25          // Output Relay 1 (Green Wire)
#define R2 26          // Output Relay 2 (Yellow Wire)
#define R3 27          // Output Relay 3 (Orange Wire)
#define R4 12          // Output Relay 4 (Brown Wire)
#define DIPSW1 17      // Dip Switch 1
#define DIPSW2 16      // Dip Switch 2
#define GSMLED 32      // Config LED ?
#define STATUSLED 13   // Status LED
#define VIN 33         // Analog pin for Voltage Sensing
#define V_DIV 242.0    // Voltage Divisor for VIN Measurement
#define V_ADD 0.75     // Voltage Additive for VIN Measurement
GSMSMS gsmsms(ESP324G);
#endif

// Prep for MEGGaBoard
#if defined(__AVR_ATmega2560__)
#define T1 4           // Trigger 1 (White Wire)
#define T2 5           // Trigger 2 (Grey Wire)
#define T3 6           // Trigger 3 (Purple Wire)
#define AD 7           // Arm/Disarm (Blue Wire)
#define R1 8           // Output Relay 1 (Green Wire)
#define R2 9           // Output Relay 2 (Yellow Wire)
#define R3 10          // Output Relay 3 (Orange Wire)
#define R4 11          // Output Relay 4 (Brown Wire)
#define DIPSW1 3       // Dip Switch 1
#define DIPSW2 2       // Dip Switch 2
#define GSMLED 12      // Config LED ?
#define STATUSLED 13   // Status LED
#define VIN A0         // Analog pin for Voltage Sensing
#define V_DIV 2.0      // Voltage Divisor for VIN Measurement
#define V_ADD 1.0      // Voltage Additive for VIN Measurement
GSMSMS gsmsms(MEGGABOARD3G);
#endif

// Globals ------------------------------------------------------------------------------------------------------------------------------------------

// Globals from EEPROM
char AlarmID[20];                 // Name of Alarm
char MasterPwd[20];               // Alarm Password
int UseGPS;                       // GPS On (1) or Off (0)
int ArmDelay;                     // Seconds to delay arming (if appropriate) after cold start
int SMSInterval;                  // Minutes between SMS notifocations
float VMin;                       // Low Voltage Alert Threshold
int Armed;                        // Alarm is Armed (0=No, 1-Yes)

char IMEI[20];                    // IMEI (Handset)
char ICCID[20];                   // ICCID (SIM)
char Carrier[40];                 // Network Operator
char NetMode[40];                 // Network Mode
char SvcCntr[20];                 // Service Center

int GSMDisabled;                  // GSM Disabled (No SIM, etc)
int GSMReady;                     // GSM is Ready
int CanNotify;                    // Have at least one Notification User Identified

int PStat;                        // Alarm Triggered (0=NO or 1,2,3,4 for Current Priority)
int NStat;                        // Notification Status (Number of users in notification list)

int ADCircuit;                    // Arm/Disarm Circuit - Only effects arm/disarm setting when circuit state CHANGES

int DipSW1;                       // State of Dip Switch 1 (AutoArm for standalone units)
int DipSW2;                       // State of Dip Switch 2 (Disable GSM for standalone units)

int CSQ;                          // Signal Quality

int Output[4];                    // Output/Relay Control/Status
int Relay[4];                     // Relay/Pin Enumeration

BLINKSTRUCT StatusLED;            // Status LED Control Structure
BLINKSTRUCT GsmLED;               // GSM LED Control Structure

char Buffer[BUFFSIZE];
char Caller[NUMSIZE];
char Temp[BUFFSIZE];

char Latitude[20];
char Longitude[20];
char ActiveTrigger[20];

unsigned long FlashTimer;
unsigned long ResetTimer;
unsigned long ResetPeriod;
unsigned long SMSTimer;
unsigned long SMSPeriod;

unsigned long MillisNow;
unsigned long LastTick;
char TimeNow[15];
char HBTime[9];

// ----------------------------  SETUP Functions  -------------------------------------

void setup() {
int i;

#if defined(ESP32)
  EEPROM.begin((EENOTIFY + EEMAXSMS) * EEFIELDSIZE);
#endif

  Serial.begin(9600);
  Serial.println();
  Serial.println();

  Serial.print("SuperCell ");
  Serial.print(VERSION);
  
  Serial.println();
  Serial.println();

  SetupPins();
  
  RTBlinker(&StatusLED, STATUSLED, 500);
  RTBlinker(&GsmLED, GSMLED, 500);

  Serial.print("EEPROM ");
  if(IsEEPROMValid()) Serial.println("Valid");
  else {
    InitializeEEPROM();
    Serial.println("Initialized");
  }
  ReadEEPROMData();

  Serial.println("");
  Serial.print("DIPSW1: ");
  if(digitalRead(DIPSW1) == HIGH) {        
    Serial.println("ON (GSM Enabled)");
    DipSW1=1;
  } else {
    Serial.println("OFF (GSM Disabled)");         // GSM Disabled, Standalone, 
    DipSW1=0;
    Armed=1;                                      // Auto Arm Standalone Units
    GSMDisabled=1;
    SetGSMLED(0);                                 // LED Off for GSM Manually Disabled
    GSMReady=0;
    UseGPS=0;
    CanNotify=0;
  }
  
  Serial.print("DIPSW2: ");
  if(digitalRead(DIPSW2) == HIGH) {        
    Serial.println("ON (Factory Reset)");
    DipSW2=1;
    InitializeEEPROM();
  } else {
    Serial.println("OFF");
    DipSW2=0;
  }
  
  Serial.print("Arm/Disarm Circuit: ");
  if(digitalRead(AD) == HIGH) {
    ADCircuit=1;
    Serial.println("ON");
  } else {
    ADCircuit=0;
    Serial.println("OFF");
  }

  DisplayConfig();

  if(DipSW1 == 1) {
    Serial.println("GSM Init...");
    gsmsms.Init();
    Serial.println("GSM SetParams...");
    gsmsms.SetParams();
    if(gsmsms.SIMInserted()) {
      gsmsms.SendCmd("AT+CNMI=2,1");              // Enable Unsolicited SMS Notification
      gsmsms.SendCmd("AT+AUTOCSQ=1,1");           // Enable Unsolicited CSQ Notification (whenever it changes)
//    gsmsms.SendCmd("AT+AUTOCSQ=1,0");           // Enable Unsolicited CSQ Notification (every 5 secs)
      Serial.println("GSM DeleteAllSMS...");
      gsmsms.DeleteAllSMS();
      if(UseGPS) {
        gsmsms.EnableGPS();
        Serial.println("GPS Started");
      }    
      GSMDisabled=0;
      for(i=0; i<EEMAXSMS; i++) if(strlen(ReadEEPROMSMSNumber(i+1,Temp))) CanNotify=1;
      CSQ=gsmsms.GetSignalQuality();
      if(CSQ == 0) {
        SetGSMLED(-1);                             // GSM Searching (Poor Signal) = Blinking
        GSMReady=0;     
        Serial.println("GSM Searching (Poor Signal)");
      } else {
        SetGSMLED(-2);                            // GSM Ready = Solid On
        GSMReady=1;
        Serial.println("GSM Ready!\n");
        if(!CanNotify) {
          Serial.println("Notification List Empty\n");
          SetGSMLED(2);                          // Unable to Notify = 2 Flashes       
        }
      }
    } else {
      GSMDisabled=1;
      SetGSMLED(3);                              // No SIM Detected = 3 Flashes!
      GSMReady=0;
      CSQ=0;
      UseGPS=0;
      CanNotify=0;
      Serial.println("GSM Unavailable (No SIM Detected)\n");
    }
  }

  DisplayNetInfo();

  NStat=0;
  for(i=0; i<EEMAXSMS; i++) {
    if(strlen(ReadEEPROMSMSNumber(i+1,Temp))) NStat++;
  } 

  PStat=0;
  strcpy(ActiveTrigger,"");
  for(i=0; i<4; i++) Output[i]=0;
  Relay[0]=R1;
  Relay[1]=R2;
  Relay[2]=R3;
  Relay[3]=R4;
  
  MillisNow=millis();
  FlashTimer=MillisNow;
  ResetTimer=(unsigned long)0;
  ResetPeriod=(unsigned long)0;
  SMSTimer=(unsigned long)0;
  SMSPeriod=(unsigned long)0;
  
  LastTick=MillisNow;
  gsmsms.GetTime(TimeNow);
  strcpy(HBTime, "01,12:00");

  if(Armed) {
    if(ArmDelay > 0) {
      Serial.print("Arm Delay ");
      Serial.print(ArmDelay);
      Serial.println(" seconds...");
      delay(ArmDelay * 1000);
    }
    ArmAlarm(1);
    RTBlinker(&StatusLED, -1);
  } else {
    DisarmAlarm(1);
    RTBlinker(&StatusLED, -2);
  }

  Serial.println("SuperCell Ready!");
  Serial.println("-------------------------------");
  Serial.println();
}

void SetupPins() {
  
  // Triggers
  pinMode(T1, INPUT);
  pinMode(T2, INPUT);
  pinMode(T3, INPUT);

  // Arm/Disarm Circuit
  pinMode(AD, INPUT);

  // Output Relays
  pinMode(R1, OUTPUT);
  pinMode(R2, OUTPUT);
  pinMode(R3, OUTPUT);
  pinMode(R4, OUTPUT);

  // DIP Switches
  pinMode(DIPSW1, INPUT);
  pinMode(DIPSW2, INPUT);
  
  // LEDs
  pinMode(STATUSLED, OUTPUT);
  pinMode(GSMLED, OUTPUT);

  // Voltage Input
  pinMode(VIN, INPUT);

  // Set Output Relays Open
  digitalWrite(R1,LOW);
  digitalWrite(R2,LOW);
  digitalWrite(R3,LOW);
  digitalWrite(R4,LOW);
  
}

void DisplayConfig() {
int i;

  Serial.println("");

  Serial.print("AlarmID: ");
  Serial.println(AlarmID);

  Serial.print("Pwd: ");
  if(strlen(MasterPwd)) Serial.println(MasterPwd);
  else Serial.println("{not set}");

  Serial.print("UseGPS: ");
  Serial.println(UseGPS);

  Serial.print("Armed: ");
  Serial.println(Armed);

  Serial.print("ArmDelay: ");
  Serial.println(ArmDelay);

  Serial.print("SMSInterval: ");
  Serial.println(SMSInterval);

  Serial.print("VMin: ");
  Serial.println(VMin);

  for(i=1; i<=3; i++) Serial.println(ReadTrigger(i, Temp));

  for(i=1; i<=4; i++) Serial.println(ReadOutput(i, Temp));

  Serial.print("Chirp: ");
  Serial.println(ReadEEPROMField(EECHIRP,Temp));
  
  for(i=0; i<EEMAXSMS; i++) {
    sprintf(Buffer,"SMSNotify %d: %s",i+1,ReadEEPROMSMSNumber(i+1,Temp));
    Serial.println(Buffer);
  }

  Serial.println("");
}

void DisplayNetInfo() {

  gsmsms.GetIMEI(IMEI);
  Serial.print("IMEI: ");
  Serial.println(IMEI);

  gsmsms.GetICCID(ICCID);
  Serial.print("ICCID: ");
  Serial.println(ICCID);

  gsmsms.GetNetworkOperator(Carrier);
  Serial.print("Carrier: ");
  Serial.println(Carrier);

  gsmsms.GetNetworkMode(NetMode);
  Serial.print("Network Mode: ");
  Serial.println(NetMode);

  gsmsms.GetServiceCenter(SvcCntr);
  Serial.print("Service Center: ");
  Serial.println(SvcCntr);
  
/*
Serial.print("Preferred Mode: ");
Serial.println(gsmsms.GetPreferredMode());
gsmsms.SetPreferredMode(13);
Serial.print("Preferred Mode Now: ");
Serial.println(GetPreferredMode());
*/
}

// ----------------------------  LOOP Functions -------------------------------------

void loop() {
int tick;

  MillisNow = millis();                   // Several Functions use this global variable!

  if(Armed) {
    CheckTriggers();
    SetRelays();
    CheckNotify();
    CheckReset();
  }

  CheckADC();

  if(!GSMDisabled) {

    if (MillisNow > (LastTick + 30000)) {   // "LastTick" every 30 secs to get <1 min time resolution
      gsmsms.GetTime(TimeNow);              // ...as this takes time!
      CheckHeartbeat();
      CheckSMS();                           // ...Just in case we missed anything
      CSQ = gsmsms.GetSignalQuality();
      LastTick = MillisNow;
    }
        
    ProcessUnsolicitedMsg();
    
    CheckGSM(NULL);                         // Just update the GSMLED
    RTBlinker(&GsmLED); 
  }

  CheckStatus();
  
  RTBlinker(&StatusLED);                    // Service the STATUS LED
  RTBlinker(&GsmLED);                       // Service the GSM LED  
}

int CheckHeartbeat() {
static int HBSent = 0;

  if(HBTime[0] == '\0') return 0;

  if(strncmp(TimeNow + 6, HBTime, strlen(HBTime)) == 0) {
    if(HBSent == 0) {
      if(strlen(ReadEEPROMSMSNumber(1,Temp)) > 0) SendStatus(Temp,"Monthly Heartbeat!");
      HBSent = 1;
      return 1;
    }
    return 0;
  }
  
  HBSent = 0;
  return 0;
}
  
void CheckADC() {
  switch (ADCircuit) {
    case 0:                                 // Was Last Off
      if(digitalRead(AD) == LOW) return;    // Still Off = No State Change
      else {                                // Just Turned ON! (State Change!)
        Serial.println("Arm/Disarm Circuit: ON");
        ADCircuit=1;
        ArmAlarm(0);
      }
      break;
    case 1:                                 // Was Last On
      if(digitalRead(AD) == HIGH) return;   // Still On = No State Change
      else {                                // Just Turned OFF! (State Change!)
        Serial.println("Arm/Disarm Circuit: OFF");
        ADCircuit=0;
        DisarmAlarm(0);
      }
      break;
  }
}

void CheckTriggers() {

  if(Armed) {

    if(digitalRead(T1) == HIGH) SetAlarm(EET1NAME, EET1INFO);
    if(digitalRead(T2) == HIGH) SetAlarm(EET2NAME, EET2INFO);
    if(digitalRead(T3) == HIGH) SetAlarm(EET3NAME, EET3INFO);
  }
}

void SetAlarm(int TName, int TInfo) {
int priority;
int reset;
char info[20];
int i;

  ReadEEPROMField(TName, ActiveTrigger);
  ReadEEPROMField(TInfo, info);

  priority = (int)info[0] - 48;
  if(priority == 0) return;

  if(PStat != 0) {
    if(priority > PStat) return;
  }
  
  ResetTimer=MillisNow;
  SMSTimer=MillisNow;
  SMSPeriod=(unsigned long)SMSInterval * TMULTIPLIER;

  if(priority == PStat) return;

  PStat = priority;
  if(GSMReady) SendAlarmNotification();
  else {
    Serial.print("Alarm Triggered: ");
    Serial.println(ActiveTrigger);
  }

  for(i=2; i<strlen(info); i++) if(info[i] == ',') {
    info[i]='\0';
    ResetPeriod=atol(info+i+1) * TMULTIPLIER;
  }

  for(i=0; i<4; i++) {
    if(strchr(info+2,(char)(i+48+1))) Output[i]=ReadEEPROMIntField(EER1ACTION+(i*2)); 
    else Output[i]=0;
  }
}

void SetRelays() {
int i;
int intervals;
unsigned long flashperiod=500;

  for(i=0; i<4; i++) {
    switch (Output[i]) {
      case 0:
         digitalWrite(Relay[i],LOW);
         break;
      case 1:
         digitalWrite(Relay[i],HIGH);
         break;
      case 2:
        intervals = (int)(MillisNow/flashperiod);                                                // Elapsed Time Intervals
        if(intervals % 2) {
          digitalWrite(Relay[i],LOW);                                                       // Odd (Off)
        } else {
          digitalWrite(Relay[i],HIGH);                                                      // Even (On)    
        }
        break;
    }
  }
}

void CheckNotify() {
int i;

  if(ResetTimer == 0l) return;                  // ResetTimer will be 0 if alarm not activated
  
  if(TimesUp(SMSTimer,SMSPeriod)) {
    if(GSMReady) SendAlarmNotification();
    SMSTimer=(MillisNow);
  }
}

void CheckReset() {

  if(ResetPeriod == 0l) return;
  
  if(TimesUp(ResetTimer,ResetPeriod)) ResetAlarm(0);
}

void CheckStatus() {
float VNow;
unsigned long VDuration = 60000;      // Min Duration of Low Voltage Event which will trigger an alarm (ms)
static int VStat = 1;                 // Voltage Status (-1=LOW, 0=Trending Low, 1=OK)
static unsigned long VTimer;
int LEDCode;

  if(Armed) LEDCode = -1;             // Armed = Flashing
  else LEDCode = -2;                  // Disarmed = Solid On

  if(VNow != 0.0) {                    // Ignore Voltage Check

    VNow = GetVin(1, 0);

    switch (VStat) {
      case -1:                        // LOW!
        if(VNow >= VMin) {
          VStat = 1;
        } else LEDCode = 2;
        break;
      case 0:                         // Trending Low...
        if(VNow >= VMin) VStat = 1;
        else if(millis() > (VTimer + VDuration)) {
          VStat = -1;
          if(GSMReady) {
            strcpy(Buffer, "Alarm Alert: Low Battery! (");
            dtostrf(VNow, 4, 1, &Buffer[strlen(Buffer)]);
            strcat(Buffer, "v)");
            SendAlarmSMS(Buffer);
          }
        }
        break;
      case 1:                       // OK
        if(VNow < VMin) {
          VStat = 0;
          VTimer = millis();
        }
        break;
    }
  }

  SetStatusLED(LEDCode);

  return;
}

void SetStatusLED(int blinks){
// Solid    = On/Disarmed (-2)
// 1 Blink  = On/Armed (-1)
// 2 Blinks  = Low Voltage

  if(StatusLED.blinks == blinks) return;      // No Change;

  RTBlinker(&StatusLED, blinks);  
}

void CheckGSM(char *p) {
char *p1;
char *p2;
int csq;

  if(p != NULL) {
    p1=strstr(Buffer,"+CSQ:");
    if(p1 == NULL) return;
    p1+=6;
  
    p2=strchr(p1,',');
    if(p2 == NULL) return;
    *p2='\0';

    csq=atoi(p1);
    
    CSQ = 0;
    if(csq > 1)  CSQ = 1;
    if(csq > 5)  CSQ = 2;
    if(csq > 9)  CSQ = 3;
    if(csq > 14) CSQ = 4;
    if(csq > 19) CSQ = 5;
    if(csq == 99) CSQ=0;
  
    if(CSQ > 0) GSMReady = 1;
    else GSMReady = 0;
  }

  if(GSMReady) {
    if(CanNotify == 0) SetGSMLED(2);
    else SetGSMLED(-2);
  } else SetGSMLED(-1);
}

void SetGSMLED(int blinks){
// Off      = GSM Disabled (DipSW1 = Off)
// Solid    = On/Normal  (-2)
// 1 Blink  = GSM Seaching (Temp/Isolated) (-1)
// 2 Blinks = No One to Nofify (Empty List)
// 3 Blinks = No SIM Detected (Fatal/Isolated)

  if(GsmLED.blinks == blinks) return;                           // No Change;

  if(GsmLED.blinks == 3) return;                                // No Recovery if No SIM

  RTBlinker(&GsmLED, blinks);
}

void ArmAlarm(int boot) {
  DoChirp(1);

  if(boot) sprintf(Buffer,"%s Alarm is Up (ARMED)", AlarmID);
  else sprintf(Buffer,"%s Alarm ARMED", AlarmID);

  if(GSMReady) SendAlarmSMS(Buffer);
  Serial.println(Buffer);

  Armed=1;
  WriteEEPROMIntField(EEARMED,Armed);
}

void DisarmAlarm(int boot) {
int i;

  Armed=0;
  WriteEEPROMIntField(EEARMED,Armed);

  if(PStat) ResetAlarm(1);

  DoChirp(2);

  if(boot) sprintf(Buffer,"%s Alarm is Up (DISARMED)", AlarmID);
  else sprintf(Buffer,"%s Alarm DISARMED", AlarmID);

  if(GSMReady) SendAlarmSMS(Buffer);
}

void ResetAlarm(int Quiet) {
int i;

  PStat=0;
  strcpy(ActiveTrigger,"");
  ResetTimer=(unsigned long)0;
  ResetPeriod=(unsigned long)0;
  SMSTimer=(unsigned long)0;
  SMSPeriod=(unsigned long)0;
  for(i=0; i<4; i++) {
    Output[i]=0;
    digitalWrite(Relay[i],LOW);
  }

  if(Quiet) return;

  sprintf(Buffer,"%s Alarm RESET", AlarmID);
  if(GSMReady) SendAlarmSMS(Buffer);
  else Serial.println(Buffer);
}

void DoChirp(int count) {
int i;
int j;

  ReadEEPROMField(EECHIRP,Temp);
  for(i=0; i<count; i++) {
    for(j=0; j<4; j++) if(strchr(Temp,j+49) != NULL) digitalWrite(Relay[j],HIGH);
    delay(CHIRPTIME);
    for(j=0; j<4; j++) if(strchr(Temp,j+49) != NULL) digitalWrite(Relay[j],LOW);
    delay(CHIRPTIME*2);
  }
}

// ----------------------------  GSM Communication Functions  -------------------------------------

void ProcessUnsolicitedMsg() {
int c;
int i;

  if(!gsmsms.Available()) return;

  Buffer[0]='\0';
  while(gsmsms.Available()) {
    c = gsmsms.Read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!gsmsms.Available()) delay(100);
  }

  Serial.print("\n\n");
  Serial.println(Buffer);   // Local Echo

  if(strstr(Buffer,"+CMTI:")) {
    CheckSMS();
    return;
  }

  if(strstr(Buffer,"+CSQ:")) {
    CheckGSM(Buffer);
    return;
  }

  return;
}

int CheckSMS() {

  if(gsmsms.GetSMS(Caller, Buffer, NUMSIZE, BUFFSIZE) == 0) return 0;

  Serial.println();
  Serial.print("(");
  Serial.print(Caller);
  Serial.print(") ");
  Serial.println(Buffer);

  ProcessCommand();

  return 1;
}

// ----------------------------  Command Processing Functions  -------------------------------------

void ProcessCommand() {
int i;
char *p1;
char *param;
int MasterUser;

  if(strlen(ReadEEPROMSMSNumber(1,Temp)) == 0) {
    if(strlen(Caller) > 4) {
      WriteEEPROMSMSNumber(1,Caller);
      CanNotify=1;
      MasterUser=1;    
    }
  } else {
    if(!strcmp(Caller,Temp)) MasterUser=1;
    else MasterUser=0;
  }

  if(!CallerAuthorized(Caller)) {
    if((strlen(Buffer)) && !strcasecmp(Buffer,MasterPwd)) {
      if(!AddSMSNotification(Caller)) {
        strcpy(Buffer,"Notification List Full! (\"CLEARNOTIFY\" to clear)");
        Serial.println(Buffer);
        gsmsms.SendSMS(Caller,Buffer);
        return;    
      } else {
        sprintf(Buffer,"%s Authorized!",Caller);
        Serial.println(Buffer);
        SendNotificationList(Caller);
        return;
      }
    } else {
      Serial.print(Caller);
      Serial.print(" Not Authorized!");
      ForwardMsgToAll();
      return;    
    }
  }

  param=strchr(Buffer,':');
  if(param != NULL) {
    if(param[1] == ' ') param += 2;
    else if(param[1] == '\0') param = NULL;
    else param += 1;
  }

  switch (GetCommandID(Buffer, CmdArray, NUMCMDS)) {

    case 0:     // Forward Message
      ForwardMsgToAll();
      return;  

    case 1:     // AlarmID:
      strncpy(AlarmID,param,20);
      AlarmID[19]='\0';
      WriteEEPROMField(EEALARMID,AlarmID);
      SendStatus(Caller, "Alarm ID Set!");
      return;

    case 2:     // UseGPS:
      if(!strcasecmp(param,"YES")) {
        SendStatus(Caller, "Starting GPS...");
        if(UseGPS == 0) gsmsms.EnableGPS();
        UseGPS=1;
      } else if(!strcasecmp(param,"NO")) {
        SendStatus(Caller, "Stopping GPS...");
        if(UseGPS == 1) gsmsms.DisableGPS();
        UseGPS=0;
      } else {
        gsmsms.SendSMS(Caller,BadCmd);
        return;          
      }
      WriteEEPROMIntField(EEUSEGPS,UseGPS);
      SendStatus(Caller, "GPS Config Updated!");
      return;

    case 3:     // SMSInterval:
      i=atoi(param);
      if(i == 0) {
        gsmsms.SendSMS(Caller,BadCmd);
        return;                  
      }
      SMSInterval=i;
      WriteEEPROMIntField(EESMSINTERVAL,SMSInterval);
      SendStatus(Caller, "SMS Interval Updated!");
      return;

    case 4:     // ArmDelay:
      for(i=0; i<strlen(param); i++) if(!isdigit(param[i])) {
        gsmsms.SendSMS(Caller,BadCmd);
        return;
      }
      ArmDelay=atoi(param);
      WriteEEPROMIntField(EEARMDELAY,ArmDelay);
      SendStatus(Caller, "Arm Delay Updated!");
      return;

    case 5:     // Trigger:
      if(!ProcessTriggerCommand(param)) SendTriggers(Caller, "Triggers Updated!");
      else SendTriggerHelp(Caller);
      return;

    case 6:     // ShowTriggers
      SendTriggers(Caller, NULL);
      return;

    case 7:     // Output:
      if(!ProcessOutputCommand(param)) SendOutputs(Caller, "Outputs Updated!");
      else SendOutputHelp(Caller);
      return;

    case 8:     // ShowOutputs
      SendOutputs(Caller, NULL);
      return;

    case 9:     // Chirp:
      if(ProcessChirpCommand(param)) gsmsms.SendSMS(Caller,BadCmd);
      else SendOutputs(Caller, "Chirp Updated!");
      return;  

    case 10:     // Notify:   
      if(MasterUser) {
        if(!SMSNumberValid(param)) gsmsms.SendSMS(Caller,"Invalid Format! Use \"+12625551212\" or \"2625551212\"");   
        else if(!AddSMSNotification(param)) gsmsms.SendSMS(Caller,"Notification List Full! (\"CLEARNOTIFY\" to clear)");
        else SendNotificationList(Caller);
        Serial.print(param);
        Serial.println(" added to Notification List");
      } else {
        strcpy(Buffer,"Not Authorized!");
        Serial.println(Buffer);
        gsmsms.SendSMS(Caller,Buffer);
      }
      return;

    case 11:     // ShowNotify
      SendNotificationList(Caller);
      return;

    case 12:     // ClearNotify
      if(MasterUser) {                    // Only First Master User Can Set PWD
        ClearSMSNotificationList();
        CanNotify=0;
        Serial.println("Notification List Cleared");
        SendNotificationList(Caller);
      } else {
        strcpy(Buffer,"Not Authorized!");
        Serial.println(Buffer);
        gsmsms.SendSMS(Caller,Buffer);
      }
      return;

    case 13:    // Arm
      ArmAlarm(0);
      // Confirmation sent to all via ArmAlarm()      
      return;

    case 14:    // Disarm
      DisarmAlarm(0);
      // Confirmation sent to all via DisarmAlarm()      
      return;

    case 15:     // Reset
      ResetAlarm(0);
      // Confirmation sent to all via ResetAlarm()      
      return;

    case 16:     // GPSLoc
      SendGPSLocation(Caller);
      return;

    case 17:     // Status
      SendStatus(Caller, NULL);
      return;

    case 18:     // NetInfo
      SendNetInfo(Caller, NULL);
      return;

    case 19:     // Config
      SendConfig(Caller, NULL);
      return;

    case 20:    // Help
      SendHelp(Caller);
      return;

    case 21:    // Pwd:
      if(MasterUser) {                    // Only First Master User Can Set PWD
        if(param != NULL) {
          WriteEEPROMField(EEPWD, param);
          strcpy(MasterPwd,param);
        } else {
          WriteEEPROMField( EEPWD, "");
          strcpy(MasterPwd,"");
        }
        strcpy(Buffer,"PWD Set!");
      } else {
        strcpy(Buffer,"Not Authorized!");
      }
      gsmsms.SendSMS(Caller,Buffer);
      Serial.println(Buffer);
      return;

    case 22:    // VMin:
      VMin=atof(param);
      WriteEEPROMField(EEVMIN,dtostrf(VMin,4,1,Temp));
      SendConfig(Caller, "VMin Updated!");
      return;

    case 23:    // ClearEEPROM (Undocumented)
      InitializeEEPROM();
      CanNotify=0;
      SendStatus(Caller, "EEPROM Initialized");
      return;

    case 24:    // SIMInfo (Undocumented)
      SendSIMInfo(Caller);
     return;

    case 25:    // MEInfo (Undocumented)
      SendMEInfo(Caller);
      return;

    case 26:    // CellInfo (Undocumented)
      SendCellInfo(Caller,gsmsms.GetSignalQuality());
      return;

    case 27:    // Operators (Undocumented)
      SendOperators(Caller);
      return;

    case 28:    // ATCMD (Undocumented)
      p1=strchr(param,'\n');
      if(p1 != NULL) *p1='\0';
      gsmsms.SendCmdParams(param,30000,1000,"OK",10);
      strncpy(Buffer,gsmsms.GSMCommBuff,BUFFSIZE);
      Buffer[BUFFSIZE-1]='\0';
      gsmsms.SendSMS(Caller,Buffer);
      return;
  }  
}

int ProcessTriggerCommand(char *p) {
// TRIGGER Configuration Fields
//
// Example: "1,1,1234,15,Door Switch"
// Field 1 = Trigger to Configure (1-4)
// Field 2 = Priority (0-4)
// Field 3 = Outputs to Trigger (1234, 134, 23, 0, etc)
// Field 4 = Reset Time {minutes}
// Filed 5 = Name

int i;
int t;
char c;
int count=0;
int len[5]={0,0,0,0,0};
char *p1;
char *p2;

  for(i=0; i<strlen(p); i++) {
    c=p[i];
    if(c == ',') {
      count++;
      if(count == 1) p1 = &p[i];
      else if(count == 4) p2 = &p[i];
    } else {
      switch (count) {
        case 0:                                             // Field 1 = Trigger # (must be 1-3)
          if((int(c) < 49) || (int(c) > 51)) return 1;
          break;
        case 1:                                             // Field 2 = Priority (must be 0-4)
          if((int(c) < 48) || (int(c) > 52)) return 2;
          break;
        case 2:                                             // Field 3 = Output List (must be 0-4)
          if((int(c) < 48) || (int(c) > 52)) return 3;
          break;
        case 3:                                             // Field 4 = Reset Time (must be digits)
          if(!isdigit(c)) return 4;
          break;
        case 4:                                             // Field 5 = Trigger Name (must be printable)
          if(!isprint(c)) return 5;
          break;
        case 5:                                             // Too Many Fields!
          return 6;
          break;
      }
      len[count]++;
    }
  }
  if(count < 4) return 7;                                 // Too Few Fields
  
  if(len[0] != 1) return 8;                               // Field 1 = Trigger # (must be one digit only)
  if(len[1] != 1) return 9;                               // Field 2 = Priority (must be one digit only)
  if((len[2] < 1) || (len[2] > 4)) return 10;             // Field 3 = Output List (must be between 1 and 4 digits long)
                                                          // Field 4 = Reset Time (No Len Check)
  if(len[4] > 19) return 11;                              // Field 5 = Trigger Name (must be 19 chars max)

  *p1 = '\0';
  p1++;
  *p2 = '\0';
  p2++;

  WriteTrigger(atoi(p), p1, p2);

  return 0;
}

int ProcessOutputCommand(char *p) {
// OUTPUT Configuration Fields
//
// Example: "1,1,Siren"
// Field 1 = Output to Configure (1-4)
// Field 2 = Function: 0=Off, 1=On, 2=Flash
// Field 3 = Name

int i;
int t;
char c;
int count=0;
int len[3]={0,0,0};
char *p1;
char *p2;

  for(i=0; i<strlen(p); i++) {
    c=p[i];
    if(c == ',') {
      count++;
      if(count == 1) p1 = &p[i];
      else if(count == 2) p2 = &p[i];
    } else {
      switch (count) {
        case 0:                                             // Field 1 = Output # (must be 1-4)
          if((int(c) < 49) || (int(c) > 52)) return 1;
          break;
        case 1:                                             // Field 2 = Action Code = OFF (0), ON (1), FLASH (2) (must be 0-2)
          if((int(c) < 48) || (int(c) > 50)) return 2;
          break;
        case 2:                                             // Field 3 = Output Name (must be printable)
          if(!isprint(c)) return 3;
          break;
        case 3:                                             // Too Many Fields!
          return 6;
          break;
      }
      len[count]++;
    }
  }
  if(count < 2) return 7;                                 // Too Few Fields

  if(len[0] != 1) return 8;                               // Field 1 = Output # (must be one digit only)
  if(len[1] != 1) return 9;                               // Field 2 = Action Code (must be one digit only)
  if(len[2] > 19) return 10;                              // Field 3 = Output Name (must be 19 chars max)

  *p1 = '\0';
  p1++;
  *p2 = '\0';
  p2++;

  WriteOutput(atoi(p), p1, p2);

  return 0;
}

int ProcessChirpCommand(char *p) {
int i;
  
  if(strlen(p) > 4) return 1;
  for(i=0; i<strlen(p); i++) {
    if((p[i] < 48) || (p[i] > 52)) return 2;
  }

  WriteEEPROMField(EECHIRP, p);
  return 0;
}

int GetCommandID(const char *astring, const char *CmdArray[], int NumCmds) {
int i;
 
  for (i=0; i<NumCmds; i++) {
    if(!strncasecmp(astring,CmdArray[i],strlen(CmdArray[i]))) return i+1;
  }
  return 0;
}

// ----------------------------  Notification Functions  -------------------------------------

int SendStatus(const char *acaller, const char *msg)
{
  Serial.print("Sending Status to: ");
  Serial.println(acaller);

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n\n");
  }

  gsmsms.Print("AlarmID:      ");
  gsmsms.Print(AlarmID);
  gsmsms.Print("\n\n");
  
  gsmsms.Print("Firmware:     ");
  gsmsms.Print(VERSION);
  gsmsms.Print("\n");

  gsmsms.Print("Signal:       ");
  gsmsms.Print(CSQ);
  gsmsms.Print(" of 5\n");

  gsmsms.Print("Volts:        ");
  gsmsms.Print(dtostrf(GetVin(1, 0),4,1,Temp));
  gsmsms.Print("\n");

  gsmsms.Print("Status:       ");
  if(PStat > 0) {
    gsmsms.Print("TRIPPED (");
    gsmsms.Print(ActiveTrigger);
    gsmsms.Print(")");
  }
  else if(Armed) gsmsms.Print("Armed");
  else gsmsms.Print("Disarmed");

  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;

  return 0;
}

int SendConfig(const char *acaller, const char *msg)
{
  Serial.print("Sending Config to: ");
  Serial.println(acaller);

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n\n");
  }

  gsmsms.Print("AlarmID:      ");
  gsmsms.Print(AlarmID);
  gsmsms.Print("\n\n");

  gsmsms.Print("UseGPS:       ");
  if(UseGPS) gsmsms.Print("Yes");
  else gsmsms.Print("No");
  gsmsms.Print("\n");

  gsmsms.Print("ArmDelay:     ");
  gsmsms.Print(ArmDelay);
  gsmsms.Print(" Sec");
  gsmsms.Print("\n");

  gsmsms.Print("SMSInterval:  ");
  gsmsms.Print(SMSInterval);
  gsmsms.Print(" Min");
  gsmsms.Print("\n");

  gsmsms.Print("VMin:  ");
  gsmsms.Print(dtostrf(VMin,4,1,Temp));
  gsmsms.Print(" Volts");

  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;

  return 0;
}

int SendNetInfo(const char *acaller, const char *msg) {
char Data[120];

  gsmsms.GetIMEI(IMEI);
  gsmsms.GetICCID(ICCID);
  gsmsms.GetNetworkOperator(Carrier);
  gsmsms.GetNetworkMode(NetMode);
  gsmsms.GetServiceCenter(SvcCntr);
  
  Serial.print("Sending Network Info to: ");
  Serial.println(acaller);

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n\n");
  }

  gsmsms.Print("IMEI:    ");
  gsmsms.Print(IMEI);
  gsmsms.Print("\n");

  gsmsms.Print("ICCID:   ");
  gsmsms.Print(ICCID);
  gsmsms.Print("\n");

  gsmsms.Print("Carrier: ");
  gsmsms.Print(Carrier);
  gsmsms.Print("\n");

  gsmsms.Print("Mode:    ");
  gsmsms.Print(NetMode);
  gsmsms.Print("\n");

  gsmsms.Print("SvcCntr: ");
  gsmsms.Print(SvcCntr);
  
  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;

  return 0;
}

void SendGPSLocation(const char *acaller) {
  Serial.print("Sending GPS Location Info to: ");
  Serial.println(acaller);

  if (UseGPS) {
    sprintf(Buffer,"%s Location:",AlarmID);
    AppendGPSInfo(Buffer);
  } else {
    sprintf(Buffer,"%s GPS Disabled!\nReply \"USEGPS: YES\" to enable.",AlarmID);
  }
  
  gsmsms.SendSMS(acaller,Buffer);
}

int SendTriggers(const char *acaller, const char *msg)
{
int  i;

  Serial.print("Sending Trigger Info to: ");
  Serial.println(acaller);

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n\n");
  }  

  gsmsms.Print("AlarmID:      ");
  gsmsms.Print(AlarmID);
  gsmsms.Print("\n\n");
  
  for(i=1; i<=3; i++) {
    gsmsms.Print(ReadTrigger(i, Temp));
    gsmsms.Print("\n");
  }

  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;

  return 0;
}

int SendOutputs(const char *acaller, const char *msg)
{
int  i;

  Serial.print("Sending Output Info to: ");
  Serial.println(acaller);

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  if(msg != NULL) {
    gsmsms.Print(msg);
    gsmsms.Print("\n\n");
  }  

  gsmsms.Print("AlarmID:      ");
  gsmsms.Print(AlarmID);
  gsmsms.Print("\n\n");
  
  for(i=1; i<=4; i++) {
    gsmsms.Print(ReadOutput(i, Temp));
    gsmsms.Print("\n");
  }

  gsmsms.Print("Chirp:");
  gsmsms.Print(ReadEEPROMField(EECHIRP,Temp));
  gsmsms.Print("\n");

  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;

  return 0;
}

int SendNotificationList(const char *acaller)
{
int  i;

  Serial.print("Sending Notification List to: ");
  Serial.println(acaller);

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  for(i=0; i<EEMAXSMS; i++) {
    gsmsms.Print("NOTIFY ");
    gsmsms.Print(i+1);
    gsmsms.Print(": ");
    gsmsms.Print(ReadEEPROMSMSNumber(i+1,Temp));
    gsmsms.Print("\n");
  }
  
  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;
}

void ForwardMsgToAll() {
int i;
int pushval;
int msglen;
int oldloc;
int newloc;

  pushval = (6 + strlen(Caller) + 2);
  Buffer[BUFFSIZE-pushval-1]='\0';
  msglen = strlen(Buffer);
  for(i=msglen; i>0; i--) {
    oldloc=i-1;
    newloc=oldloc+pushval;
    Buffer[newloc] = Buffer[oldloc];
  }
  strncpy(Buffer,"(From ",6);
  strncpy(Buffer+6,Caller,strlen(Caller));
  strncpy(Buffer+6+strlen(Caller),") ",2); 
  
  SendAlarmSMS(Buffer);
}

void SendAlarmNotification() {

  sprintf(Buffer,"%s Alarm Activated (%s)!",AlarmID,ActiveTrigger);
  if(UseGPS) AppendGPSInfo(Buffer);
  SendAlarmSMS(Buffer);
}

void SendAlarmSMS(const char *msg) {
int i;

  Serial.println();
  Serial.println(msg);
  for(i=0; i<EEMAXSMS; i++) {
    ReadEEPROMSMSNumber(i+1,Temp);
    if(strlen(Temp)) {
      Serial.print("Sending SMS To: ");
      Serial.println(Temp);
      gsmsms.SendSMS(Temp,msg);
      delay(SMSDELAY*1000);
    }
  }
  Serial.println();
}

int AppendGPSInfo(char *p) {
char *p1;

  if(!gsmsms.GetGPSPosition(Latitude, Longitude)) strcat(p,"Lat:  N/A\nLong: N/A");
  else {
    strcat(p,"\n\n");
    strcat(p,"Lat:  ");
    strcat(p,Latitude);
    strcat(p,"\nLong: ");
    strcat(p,Longitude);
    strcat(p,"\n\n");
    strcat(p,GOOGLE_MAPS_LOCATOR);
    strcat(p,Latitude);
    strcat(p,",");
    strcat(p,Longitude);
  }
  return 1;
}

// ----------------------------  Generic Tech Info/Support Functions  -------------------

void SendSIMInfo(const char *acaller){
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // ICCID
  gsmsms.GetICCID(Temp);
  strcat(Buffer,"ICCID: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMSI
  gsmsms.GetIMSI(Temp);
  strcat(Buffer,"IMSI: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Service Provider
  gsmsms.GetATResult("AT+CSPN?","+CSPN: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Provider: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Subscriber Number
  gsmsms.GetSubscriberNumber(Temp);
  strcat(Buffer,"Number: ");
  strcat(Buffer,Temp);
  
  gsmsms.SendSMS(acaller,Buffer);
}

void SendMEInfo(const char *acaller) {
  
  strcpy(Buffer,"");

  // Chipset ID
  gsmsms.GetChipset(Temp);
  strcat(Buffer,"Chipset: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMEI
  gsmsms.GetIMEI(Temp);
  strcat(Buffer,"IMEI: ");
  strcat(Buffer,Temp);

  gsmsms.SendSMS(acaller,Buffer);  
}

void SendCellInfo(const char *acaller, int csq) {
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // Current Band
  switch(gsmsms.GetSimCom()) {
    case 808:
    case 900:
      gsmsms.GetATResult("AT+CBAND?","+CBAND: ",Temp,1000);
      break;
    case 5320:
      gsmsms.GetATResult("AT*CNTI=0","*CNTI: 0,",Temp,1000);
      break;
  }
  strcat(Buffer,"Band: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Signal Quality
  sprintf(Temp,"Signal: %c/5",char(csq+48));
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
 
  // Current Network Operator and Mode
  gsmsms.GetATResult("AT+COPS?","+COPS: ",Temp,10000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Current: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Service Center Address
  gsmsms.GetATResult("AT+CSCA?","+CSCA: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"SMSC: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");

  gsmsms.SendSMS(acaller,Buffer);  
}
 
void SendOperators(const char *acaller) {

  strcpy(Buffer,"");

  // Network Operators Currently in the Network
  gsmsms.GetATResult("AT+COPS=?","+COPS: ",Temp,60000);
  strcat(Buffer,"Operators: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  gsmsms.SendSMS(acaller,Buffer);  
}

// ----------------------------  Help Message Functions  -------------------------------------

void SendHelp(const char *acaller) {
  Serial.print("Sending Help to: ");
  Serial.println(acaller);

  SendCmds1Help(acaller);
  Serial.println("Cmds1 Help Sent");
  
  delay(SMSDELAY*1000);
  SendCmds2Help(acaller);
  Serial.println("Cmds2 Help Sent");

  delay(SMSDELAY*1000);
  SendTriggerHelp(acaller);
  Serial.println("Trigger Help Sent");

  delay(SMSDELAY*1000);
  SendOutputHelp(acaller);
  Serial.println("Output Help Sent");
}

int SendCmds1Help(const char *acaller) {

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  gsmsms.Print("Commands (Page 1 of 2)\n\n");

  gsmsms.Print("ALARMID: {text}\n");
  gsmsms.Print("USEGPS: [YES|NO]\n");
  gsmsms.Print("SMSINTERVAL: {Min}\n");
  gsmsms.Print("ARMDELAY: {Sec}\n");
  gsmsms.Print("CHIRP: {outputs}\n");
  gsmsms.Print("TRIGGER: (See Below)}\n");
  gsmsms.Print("SHOWTRIGGERS\n");

  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;
}

int SendCmds2Help(const char *acaller) {

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;
  
  gsmsms.Print("Commands (Page 2 of 2)\n\n");

  gsmsms.Print("OUTPUT: (See Below)\n");
  gsmsms.Print("SHOWOUTPUTS\n");
  gsmsms.Print("NOTIFY: {SMS Number}\n");
  gsmsms.Print("SHOWNOTIFY\n");
  gsmsms.Print("CLEARNOTIFY\n");
  gsmsms.Print("ARM\n");
  gsmsms.Print("DISARM\n");
  gsmsms.Print("RESET\n");
  gsmsms.Print("GPSLOC\n");
  gsmsms.Print("STATUS");
 
  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;
}

int SendTriggerHelp(const char *acaller) {

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  gsmsms.Print("TRIGGER: {T},{P},{XXXX},{R},{Name}\n\n");

  gsmsms.Print("  {T} = Trigger\n");
  gsmsms.Print("  {P} = Priority\n");
  gsmsms.Print("  (XXXX} = Output(s)\n");
  gsmsms.Print("  {R} = Reset Mins\n");
  gsmsms.Print("  {Name} = Name");
 
  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;  
}

int SendOutputHelp(const char *acaller) {

  if(gsmsms.PrepToSendSMS(acaller) > 0) return 1;

  gsmsms.Print("OUTPUT: {X},{0|1|2}\n\n");

  gsmsms.Print("  {X} = Output To Configure (1-4)\n");
  gsmsms.Print("  (0|1|2} = 0=Off, 1=On, 2=Flash\n");
 
  gsmsms.Print(MsgEnd);
  if(gsmsms.FlushSMS() > 0) return 2;
  
  return 0;  
}

// ----------------------------  Notification List Management Functions  ---------------

int AddSMSNotification(const char *num) {
int i;
char anum[20];

  for(i=0; i<EEMAXSMS; i++) {
    ReadEEPROMSMSNumber(i+1,anum);
    if(!strcmp(anum,num)) return 2;
    if(!strlen(anum)) break;
  }
  if(i == EEMAXSMS) return 0;

  if(!strncasecmp(num,"+1",2)) WriteEEPROMSMSNumber(i+1, num);
  else {
    sprintf(anum,"+1%s",num);
    WriteEEPROMSMSNumber(i+1, anum);
  }
  
  return 1;
}

void ClearSMSNotificationList() {
int i;

  for(i=0; i<EEMAXSMS; i++) WriteEEPROMSMSNumber(i+1,"");
}

// ----------------------------  Misc Functions  -------------------------------------

int CallerAuthorized(const char *num) {
int i;

  for(i=0; i<EEMAXSMS; i++) {
    ReadEEPROMSMSNumber(i+1,Temp);
    if(!strcmp(Temp, num)) return 1;
  }
  return 0; 
}

int SMSNumberValid(const char *num) {
  
  if(strlen(num) < 10) return 0;
  if(strchr(num,'-')) return 0;

  if((strlen(num) > 10) && (num[0] != '+')) return 0;

  return 1;
}

float GetVin(int samples, int msdelay) {
float vin;

 vin = 0.0;
  for(int i=0; i<samples; i++) {
    vin += (((float)analogRead(VIN))/V_DIV) + V_ADD;
    delay(msdelay);
  }
  vin = vin / (float)samples;

  return vin;
}

int TimesUp(unsigned long timer, unsigned long timeout) {

  if(timer+timeout < timer) {     // Unsigned Long "Wrap Around" 
    if((MillisNow < (timer+timeout)) || (MillisNow > timer)) return 0;
  } else {
    if(MillisNow < (timer+timeout)) return 0;
  }
  return 1;
}

int RTBlinker(BLINKSTRUCT *blinkstruct, int p2, unsigned long msdelay) {
int blinkcount;

// if msdelay is specified (!= 0) then p2=led pin number for initialization
// else p2=blinkcount for control

  if(msdelay == 0l) blinkcount=p2;            // Control Call
  else {                                      // Initialization Call
    blinkstruct->led = p2;
    if(blinkstruct->led >= 0) pinMode(blinkstruct->led, OUTPUT);
    blinkstruct->msdelay = msdelay;
    blinkstruct->timeref = 0;
    blinkstruct->blinks = 0;
    blinkstruct->state = -1;
    return 0;
  }

// blinkcount Control:
//
//  -3 = Continue Blinking / Processing
//  -2 = Solid On (No Blinking)
//  -1 = Set Continuous Blink (no blink pattern)
//   0 = Stop Blinking
//  1+ = Set Blink Pattern

// Return Codes:
//
// Low -> High = 1  (State Change)
// High -> Low = 2  (State Change)
// High = -1        (No Change)
// Low = -2         (No Change)
// Idle = 0
// Sequence Completed = 3

int elapsed;
  
  switch(blinkcount) {
    case -3:                  // Continue Blinking/Processing
      if(blinkstruct->state < 0) return 0;
      break;
    case -2:                   // Set Solid On
      blinkstruct->blinks = blinkcount;
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
      blinkstruct->timeref = 0;
      blinkstruct->state = 1;
      return 1;
    case 0:                   // Stop Blinking / Turn Off
      blinkstruct->blinks = blinkcount;
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
      blinkstruct->timeref = 0;
      blinkstruct->state = -1;
      return 2;
    default:                  // Set Repeating Blink (>0) or Continuous Blink (-1)
      blinkstruct->blinks=blinkcount;
      blinkstruct->timeref = millis();
      blinkstruct->state=0;
      return 1;
  }

  if(blinkstruct->blinks == -2) return -1;

  elapsed = int(((millis()-blinkstruct->timeref))/blinkstruct->msdelay);      // Elapsed Time Units

  if((elapsed < (blinkstruct->blinks*2)) || (blinkstruct->blinks == -1)) {
    if (elapsed % 2) {                            // Odd (Off)
      if(blinkstruct->state == 1) {
        if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
        blinkstruct->state=0;
        return 2;
      }
      return -1;
    } else {                                      // Even (On)
      if(blinkstruct->state == 0) {
        if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
        blinkstruct->state=1;
        return 1;
      }  
      return -2;
    }
  } else {
    if(blinkstruct->state == 1) {
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, LOW);
      blinkstruct->state=0;
      return 2;
    }
    if(elapsed >= ((blinkstruct->blinks*2)+2)) {
      blinkstruct->timeref=millis();
      if(blinkstruct->led >= 0) digitalWrite(blinkstruct->led, HIGH);
      blinkstruct->state=1;
      return 3;
    } 
    return -2;
  }
}

// ----------------------------  EEPROM Functions  -------------------------------------

void WriteTrigger(int n, const char *tinfo, const char *tname) {
  WriteEEPROMField(EET1INFO+((n-1)*2),tinfo);
  WriteEEPROMField(EET1NAME+((n-1)*2),tname);
}

void WriteOutput(int n, const char *raction, const char *rname) {
  WriteEEPROMField(EER1ACTION+((n-1)*2),raction);
  WriteEEPROMField(EER1NAME+((n-1)*2),rname);
}

char *ReadTrigger(int n, char *s) {
char tinfo[20];
char tname[20];

  ReadEEPROMField(EET1INFO+((n-1)*2), tinfo);
  ReadEEPROMField(EET1NAME+((n-1)*2), tname);
  sprintf(s,"Trigger:%d,%s,%s", n, tinfo, tname);
  return s;
}

char *ReadOutput(int n, char *s) {
char raction[20];
char rname[20];

  ReadEEPROMField(EER1ACTION+((n-1)*2), raction);
  ReadEEPROMField(EER1NAME+((n-1)*2), rname);
  sprintf(s,"Output:%d,%s,%s", n, raction, rname);
  return s;
}

void ReadEEPROMData() {
  ReadEEPROMField(EEALARMID,AlarmID);
  ReadEEPROMField(EEPWD,MasterPwd);
  UseGPS=ReadEEPROMIntField(EEUSEGPS);
  Armed=ReadEEPROMIntField(EEARMED);
  ArmDelay=ReadEEPROMIntField(EEARMDELAY);
  SMSInterval=ReadEEPROMIntField(EESMSINTERVAL);
  VMin=atof(ReadEEPROMField(EEVMIN,Temp));
}

void WriteEEPROMData() {
  WriteEEPROMField(EEALARMID,AlarmID);
  WriteEEPROMField(EEPWD,MasterPwd);
  WriteEEPROMIntField(EEUSEGPS,UseGPS);
  WriteEEPROMIntField(EEARMED, Armed);
  WriteEEPROMIntField(EEARMDELAY, ArmDelay);
  WriteEEPROMIntField(EESMSINTERVAL,SMSInterval);
  WriteEEPROMField(EEVMIN,dtostrf(VMin,4,1,Temp));
}

int IsEEPROMValid() {

  ReadEEPROMField(EESIG,Buffer);
  if(!strcmp(Buffer, EEPROMSig)) return 1;
  return 0;
}

void InitializeEEPROM() {
int i;
  
  WriteEEPROMField(EESIG, EEPROMSig);
  
  // Defaults Settings
  strcpy(AlarmID,"SuperCell");
  strcpy(MasterPwd,"");
  UseGPS=0;
  Armed=0;
  ArmDelay=10;
  SMSInterval=5;
  VMin=11.0;
  WriteEEPROMData();

  // EEPROM Trigger Configuration Fields (20 Bytes)
  //
  // Example: "1,1234,15"
  // Field 1 = Priority (0-4)
  // Field 2 = Outputs to Trigger (1234, 134, 23, 0, etc)
  // Field 3 = Reset Time {minutes}

  WriteEEPROMField(EET1NAME, "Door Switch");
  WriteEEPROMField(EET1INFO, "1,1234,5");
  
  WriteEEPROMField(EET2NAME, "PIR Motion");
  WriteEEPROMField(EET2INFO, "1,1234,5");
  
  WriteEEPROMField(EET3NAME, "Trailer Motion");
  WriteEEPROMField(EET3INFO, "1,1234,5");
  
  WriteEEPROMField(EER1ACTION, "1");
  WriteEEPROMField(EER1NAME,   "Siren");

  WriteEEPROMField(EER2ACTION, "2");
  WriteEEPROMField(EER2NAME,   "Marker Lights");

  WriteEEPROMField(EER3ACTION, "1");
  WriteEEPROMField(EER3NAME,   "Brakes");

  WriteEEPROMField(EER4ACTION, "1");
  WriteEEPROMField(EER4NAME,   "Interior Light");

  WriteEEPROMField(EECHIRP, "12");

  for(i=0; i<EEMAXSMS; i++) WriteEEPROMSMSNumber(i+1,"");
}

int WriteEEPROMSMSNumber(int index,const char *text){
  return WriteEEPROMField(EENOTIFY+index-1, text);
}

int WriteEEPROMField(int field, const char *text) {
  return WriteEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
}

int WriteEEPROMIntField(int field, int i) {
char text[EEFIELDSIZE];

  return WriteEEPROMField(field, itoa(i,text,10));
}

int WriteEEPROMString(int offset, const char *text, int maxlen) {
int i;
int c;

  if((offset+maxlen) >= EEPROM_BYTES) return 0;

  for (i=0; i<maxlen; i++) {
    if(i == (maxlen-1)) EEPROM.write(offset+i, '\0');
    else EEPROM.write(offset+i, text[i]);
    if(text[i] == '\0') break;
  }

#if defined(ESP32)
  EEPROM.commit();
#endif
  
  return i+1;
}

char *ReadEEPROMSMSNumber(int index,char *text){
  ReadEEPROMField(EENOTIFY+index-1, text);
  return text;
}

char *ReadEEPROMField(int field, char *text) {
  strcpy(text,"");
  ReadEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
  return text;
}

int ReadEEPROMIntField(int field) {
char text[EEFIELDSIZE];

  ReadEEPROMField(field, text);
  return atoi(text);
}

int ReadEEPROMString(int offset, char *text, int maxlen) {
int i;

  if((offset+maxlen) >= EEPROM_BYTES) return 0;

  for (i=0; i<maxlen; i++) {
    text[i] = EEPROM.read(offset+i);
    if(text[i] == '\0') break;
  }
  
  text[maxlen-1]='\0';

  return i+1;
}
